/* NetworkManager system settings service
 *
 * Mikhail Efremov <sem@altlinux.org>
 *
 * Based on ifcfg-fedora and ifcfg-suse code
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2008 - 2012 Mikhail Efremov
 */

#include "common.h"
#include <gmodule.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#include <string.h>
#include <sys/inotify.h>
#include <unistd.h>
#include <errno.h>
#include <net/ethernet.h>
#include <netinet/ether.h>
#include <gio/gio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <stdlib.h>

#include <nm-settings-connection.h>
#include <nm-setting-wired.h>
#include <nm-setting-wireless.h>
#include <nm-setting-gsm.h>
#include <nm-setting-cdma.h>
#include <nm-setting-pppoe.h>
#include <nm-setting-wireless-security.h>
#include <nm-setting-8021x.h>
/* Hacks necessary to #include wireless.h; yay for WEXT */
#ifndef __user
#define __user
#endif
#include <sys/types.h>
#include <linux/types.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <linux/wireless.h>

#define G_UDEV_API_IS_SUBJECT_TO_CHANGE
#include <gudev/gudev.h>

#include "nm-dbus-glib-types.h"
#include "plugin.h"
#include "reader.h"
#include "shvar.h"
#include "nm-etcnet-connection.h"

#include <arpa/inet.h>

#include <stdio.h>

#define DEFAULT_PROFILE_KEY "#"

static void system_config_interface_init (NMSystemConfigInterface *system_config_interface_class);

G_DEFINE_TYPE_EXTENDED (SCPluginIfcfg, sc_plugin_ifcfg, G_TYPE_OBJECT, 0,
						G_IMPLEMENT_INTERFACE (NM_TYPE_SYSTEM_CONFIG_INTERFACE,
											   system_config_interface_init))

#define SC_PLUGIN_IFCFG_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SC_TYPE_PLUGIN_IFCFG, SCPluginIfcfgPrivate))

#define ETCNET_PATH SYSCONFDIR"/net"
#define ETCNET_IFACES_PATH ETCNET_PATH"/ifaces/"

typedef struct
{
	GFileMonitor *monitor;
	guint id;
}FileMonitor;

typedef struct
{
	SCPluginIfcfg *plugin;

	char *iface_path;
	char *mac;
	gboolean unmanaged;
	NMDeviceType dev_type;
	GHashTable *connections;

	FileMonitor iface_monitor;
}NMEtcnetDevice;

typedef struct
{
	GUdevClient *client;

	GHashTable *devices;
	char *ifaces_path;
	FileMonitor ifaces_monitor;

	char *hostname;
	FileMonitor network_file_monitor;

	char *current_profile;
} SCPluginIfcfgPrivate;

static void update_device_config(NMEtcnetDevice *etcnet_device);

static void
ignore_cb (NMSettingsConnection *connection,
        GError *error,
        gpointer user_data)
{
}

static char *
get_iface_path(const char *ifaces_path, const char *iface)
{
	return g_build_filename (ifaces_path, iface, NULL);
}

static const char *
iface_config_files[]=
{
	"options",
	"ipv4address",
	"ipv4route",
	"ipv6address",
	"ipv6route",
	"iplink",
	"wpa_supplicant.conf",
	NULL
};

static void
start_monitor (FileMonitor *file_monitor, const char *filepath, GCallback cb_func, gpointer user_data)
{
	GFile *file;
	GFileMonitor *monitor;

	g_return_if_fail(file_monitor != NULL);
	g_return_if_fail(filepath != NULL);

	file = g_file_new_for_path (filepath);
	monitor = g_file_monitor (file, G_FILE_MONITOR_NONE, NULL, NULL);
	g_object_unref (file);

	if (monitor) {
		file_monitor->id = g_signal_connect (monitor, "changed", cb_func, user_data);
		file_monitor->monitor = monitor;
	}
}

static void
stop_monitor (FileMonitor *file_monitor)
{
	g_return_if_fail(file_monitor != NULL);

	if (file_monitor->monitor) {
		if (file_monitor->id)
			g_signal_handler_disconnect (file_monitor->monitor, file_monitor->id);

		g_file_monitor_cancel (file_monitor->monitor);
		g_object_unref (file_monitor->monitor);

		file_monitor->id = 0;
		file_monitor->monitor = NULL;
	}
}

#define NETPROFILE_OPTION "netprofile"

/* TODO: Call selectprofile script for each iface */
static char *
get_current_profile (void)
{
	char *line = NULL;
	char *profile = NULL;
	gsize line_len = 0;
	GError *error = NULL;

	profile = getenv ("NETPROFILE");
	if (profile)
		return g_strdup (profile);

	if (g_file_get_contents ("/proc/cmdline", &line, &line_len, &error)) {
		char *b, *e;

		if (line[line_len - 1] == '\n')
			line[line_len - 1] = '\0';
		b = strstr (line, NETPROFILE_OPTION"=");
		if (b) {
			b += strlen (NETPROFILE_OPTION"=");
			e = strpbrk (b, " \t");
			if(e)
				*e = '\0';
			profile = g_strdup (b);
		}
	} else
		PLUGIN_ERROR ("Couldn't read /proc/cmdline", error);


	g_free (line);

	return profile ?: g_strdup ("#");
}

static void
get_unmanaged_devices_cb (gpointer key, gpointer val, gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) val;
	GSList **list = (GSList **) user_data;

	if(etcnet_device->unmanaged) {
		char *spec = g_strdup_printf ("mac:%s", etcnet_device->mac);
		*list = g_slist_prepend (*list, spec);
	}
}

static GSList *
get_unmanaged_specs (NMSystemConfigInterface *config)
{
	GSList *list = NULL;

	g_hash_table_foreach (SC_PLUGIN_IFCFG_GET_PRIVATE (config)->devices,
					  get_unmanaged_devices_cb, &list);
	return list;
}

static gboolean
is_file_ignored(const char *filename)
{
	const char **fname;
	size_t fname_len;

	g_return_val_if_fail (filename != NULL, TRUE);

	if(filename[0] == '#' || filename[0] == '.' ||
		    g_str_has_suffix(filename, TILDE_TAG) ||
		    g_str_has_suffix(filename, BAK_TAG))
		return TRUE;

	fname_len = strlen (filename);
	for(fname = iface_config_files; *fname != NULL; ++fname) {
		size_t len = strlen(*fname);
		if(len <= fname_len &&
			    !strncmp(*fname, filename, len) &&
			    /* Exactly match or profiled name */
			    (filename[len] == '\0' || filename[len] == '#'))
			return FALSE;
	}

	return TRUE;
}

static char *
get_profile(const char *filename)
{
	char *p;

	g_return_val_if_fail (filename != NULL, NULL);

	p = strchr(filename, '#');
	if(p != NULL)
		return g_strdup(p + 1);

	return NULL;
}

static char *
find_profile(GDir *iface_dir)
{
	const char *file_name;
	g_return_val_if_fail (iface_dir != NULL, NULL);

	while((file_name = g_dir_read_name(iface_dir)))	{
		char *profile;
		if(is_file_ignored(file_name)) {
			continue;
		}
		profile = get_profile(file_name);
		if(profile != NULL)
			return profile;
	}

	return NULL;
}

static gint
compare_str(gconstpointer a, gconstpointer b)
{
	return strcmp(a, b);
}

static GSList *
get_all_profiles(const char *iface_path)
{
	GDir *iface_dir;
	GSList *profiles_list = NULL;
	char *profile = NULL;

	iface_dir = g_dir_open(iface_path, 0, NULL);
	if (!iface_dir)	{
		return NULL;
	}

	while((profile = find_profile(iface_dir))) {
		if(!g_slist_find_custom(profiles_list, profile, compare_str)) {
			profiles_list = g_slist_prepend(profiles_list, profile);
		} else {
			g_free(profile);
		}
	}

	return profiles_list;
}

static gboolean
is_profile_equal(gconstpointer a, gconstpointer b)
{
	const char *pfl1 = (const char *)a;
	const char *pfl2 = (const char *)b;

	if(!pfl1 && !pfl2)
		return TRUE;
	if(pfl1 && pfl2)
		return strcmp(pfl1, pfl2) ? FALSE : TRUE;
	return FALSE;
}

static char *
get_hostname (void)
{
	shvarFile *network;
	char *hostname;

	network = svNewFile (NETWORK_FILE);
	if (!network) {
		PLUGIN_WARNING ("Could not get hostname: failed to read " NETWORK_FILE);
		return FALSE;
	}

	hostname = svGetValue (network, "HOSTNAME");
	/* Ignore a hostname of 'localhost' or 'localhost.localdomain' */
	if (hostname && (!strcmp (hostname, "localhost") || !strcmp (hostname, "localhost.localdomain"))) {
		g_free (hostname);
		hostname = NULL;
	}

	svCloseFile (network);
	return hostname;
}

static void reset_device_connections(NMEtcnetDevice *etcnet_device);

static const char *
etcnet_dev_get_iface_name (const NMEtcnetDevice* etcnet_device)
{
	const char *p = NULL;
	g_return_val_if_fail (etcnet_device != NULL, NULL);

	if (etcnet_device->iface_path) {
		p = strrchr (etcnet_device->iface_path, '/');
		if (p)
			++p;
	}

	return p;
}

static void
device_dir_changed (GFileMonitor *monitor,
		   GFile *file,
		   GFile *other_file,
		   GFileMonitorEvent event_type,
		   gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) user_data;
	char *name;

	name = g_file_get_basename (file);
	if ( !is_file_ignored (name)) {
		switch (event_type) {
			case G_FILE_MONITOR_EVENT_DELETED:
			case G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT:
				reset_device_connections(etcnet_device);
				update_device_config(etcnet_device);
				break;
			default:
				break;
		}
	}
	g_free (name);
}

static gboolean
is_device_path_equal (gpointer key, gpointer value, gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) value;
	char *path = (char *)user_data;

	g_return_val_if_fail (etcnet_device != NULL, FALSE);
	g_return_val_if_fail (path != NULL, FALSE);

	return strcmp (etcnet_device->iface_path, path) ? FALSE : TRUE;
}

static void
ifaces_dir_changed_cb (GFileMonitor *monitor,
		   GFile *file,
		   GFile *other_file,
		   GFileMonitorEvent event_type,
		   gpointer user_data)
{
	SCPluginIfcfg *plugin = (SCPluginIfcfg *)user_data;
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);
	char *path;
	NMEtcnetDevice *etcnet_device = NULL;

	path = g_file_get_path(file);
	if(path)
		etcnet_device = g_hash_table_find (priv->devices, is_device_path_equal, path);

	switch (event_type) {
	case G_FILE_MONITOR_EVENT_DELETED:
		if(etcnet_device) {
			stop_monitor(&etcnet_device->iface_monitor);
		}
		break;
	case G_FILE_MONITOR_EVENT_CREATED:
		if(etcnet_device) {
			start_monitor(&etcnet_device->iface_monitor, etcnet_device->iface_path,
				G_CALLBACK (device_dir_changed), etcnet_device);
			reset_device_connections(etcnet_device);
			update_device_config(etcnet_device);
		}
		break;
	default:
		break;
	}

	g_free(path);
}

static void
network_file_changed_cb (GFileMonitor *monitor,
			GFile *file,
			GFile *other_file,
			GFileMonitorEvent event_type,
			gpointer user_data)
{
	SCPluginIfcfg *plugin = (SCPluginIfcfg *)user_data;
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);
	gboolean changed = FALSE;
	char *new_hostname = NULL;

	switch (event_type)	{
		case G_FILE_MONITOR_EVENT_DELETED:
			if(priv->hostname) {
				g_free(priv->hostname);
				priv->hostname = NULL;
				changed = TRUE;
			}
			break;

		case G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT:
			new_hostname = get_hostname();
			if ((new_hostname && !priv->hostname)
					|| (!new_hostname && priv->hostname)
					|| (priv->hostname && new_hostname && strcmp (priv->hostname, new_hostname))) {
				g_free (priv->hostname);
				priv->hostname = new_hostname;
				changed = TRUE;
			} else
				g_free (new_hostname);
			break;
		default:
			break;
	}

	if (changed)
		g_object_notify (G_OBJECT (plugin), NM_SYSTEM_CONFIG_INTERFACE_HOSTNAME);
}

static void
add_one_connection_for_device(NMEtcnetDevice *etcnet_device, const char *profile, GHashTable *wsconf,
				gboolean *readed)
{
	NMEtcnetConnection *connection;
	gboolean ignored;

	g_return_if_fail (etcnet_device != NULL);
	g_return_if_fail (readed != NULL);

	connection = nm_etcnet_connection_new (etcnet_device->iface_path, profile, etcnet_device->dev_type,
                                           etcnet_device->mac, wsconf, &ignored);
	if(ignored || connection)
		*readed = TRUE;
	if(connection) {
		const char *ssid = NULL;
		char *connection_key = NULL;

		if (ignored) {
			PLUGIN_WARNING ("Ignored, but readed connection");
			g_object_unref (connection);
			return;
		}
		if(profile == NULL)
			profile = DEFAULT_PROFILE_KEY;
		if(wsconf)
			ssid = (const char *)g_hash_table_lookup(wsconf, "ssid");
		connection_key = ssid ? g_strdup_printf ("%s(%s)", profile, ssid) : g_strdup (profile);
		g_hash_table_insert (etcnet_device->connections, connection_key, connection);
		PLUGIN_INFO ("%s: Readed connection \"%s\"", etcnet_dev_get_iface_name (etcnet_device),
				nm_etcnet_connection_get_id (connection));
	}
}

static void
add_connections_for_device(NMEtcnetDevice *etcnet_device, const char *profile,
				gboolean *readed)
{
	GHashTable *wsconf = NULL;
	GSList *wsconfig_list = NULL;
	GSList *list_item;
	guint num_readed = 0;
	gboolean r = FALSE;

	g_return_if_fail (readed != NULL);

	list_item = wsconfig_list = read_supplicant_config(etcnet_device->iface_path, profile);
	do {
		if(list_item)
			wsconf = (GHashTable *)list_item->data;
		add_one_connection_for_device(etcnet_device, profile, wsconf, &r);
		if (r)
			++num_readed;
		list_item = g_slist_next(list_item);
	} while(list_item);
	*readed = num_readed > 0;
	clear_supplicant_config(wsconfig_list);
}

static GSList *
get_connections_for_profile (NMEtcnetDevice *dev, const char *profile, gboolean get_default)
{
	GSList *con_list = NULL;
	GList *keys_list, *iter;
	NMEtcnetConnection *connection;
	size_t len;

	g_return_val_if_fail (dev != NULL, NULL);

	if (!profile)
		profile = DEFAULT_PROFILE_KEY;
	switch (dev->dev_type) {
		case NM_DEVICE_TYPE_ETHERNET:
			connection = (NMEtcnetConnection *) g_hash_table_lookup (dev->connections,
					profile);
			if (connection)
				con_list = g_slist_prepend (con_list, connection);
			break;

		case NM_DEVICE_TYPE_WIFI:
			keys_list = g_hash_table_get_keys (dev->connections);
			if (!keys_list)
				break;
			len = strlen (profile);
			for (iter = keys_list; iter; iter = g_list_next (iter)) {
				const char *key = (const char *)iter->data;

				if (strncmp (profile, key, len) == 0) {
					connection = (NMEtcnetConnection *) g_hash_table_lookup (dev->connections,
							key);
					if (connection)
						con_list = g_slist_prepend (con_list, connection);
				}
			}
			break;

		default:
			PLUGIN_WARNING ("Unknown device type: %d", dev->dev_type);
			break;
	}

	/* If no not-ignored connections for current profile and no was readed
	 * (no ignored connections for profile) then use connections for default
	 * profile.
	 */
	if (!con_list && get_default && strcmp (profile, DEFAULT_PROFILE_KEY))
		con_list = get_connections_for_profile (dev, NULL, FALSE);

	return con_list;
}

static gboolean
read_connections_for_one_iface(NMEtcnetDevice *etcnet_device)
{
	gboolean default_readed = FALSE;
	gboolean profiled_readed = FALSE;
	GSList *con_list;

	g_return_val_if_fail (etcnet_device != NULL, FALSE);

	if(!g_file_test(etcnet_device->iface_path, G_FILE_TEST_IS_DIR))	{
		PLUGIN_WARNING ("Couldn't access iface directory %s .", etcnet_device->iface_path);
		if(etcnet_device->dev_type != NM_DEVICE_TYPE_ETHERNET)
			return FALSE;

		add_one_connection_for_device(etcnet_device, NULL, NULL, &default_readed);
	}else {
		GSList *profiles, *iter;

		add_connections_for_device(etcnet_device, NULL, &default_readed);
		profiles = get_all_profiles(etcnet_device->iface_path);
		for (iter = profiles; iter; iter = g_slist_next (iter)) {
			char *profile_name = (char *)iter->data;

			add_connections_for_device(etcnet_device, profile_name, &profiled_readed);

			g_free(profile_name);
		}
		if(profiles)
			g_slist_free(profiles);
	}

	/* Find connections for current profile */
	con_list = get_connections_for_profile (etcnet_device,
			SC_PLUGIN_IFCFG_GET_PRIVATE (etcnet_device->plugin)->current_profile,
			!profiled_readed);
	if (con_list) {
		GSList *iter;
		NMEtcnetConnection *connection;

		etcnet_device->unmanaged = FALSE;
		for (iter = con_list; iter; iter = g_slist_next (iter)) {
			connection = (NMEtcnetConnection *)iter->data;
			if (nm_etcnet_connection_get_can_autoconnect (connection)) {
				NMSettingConnection *s_con;

				s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (NM_CONNECTION (connection),
							NM_TYPE_SETTING_CONNECTION));
				if (s_con)
					g_object_set (s_con, NM_SETTING_CONNECTION_AUTOCONNECT, TRUE, NULL);
			}
		}
	} else
		etcnet_device->unmanaged = default_readed;

	return TRUE;
}

static void
connection_added_cb(gpointer key, gpointer val, gpointer user_data)
{
	SCPluginIfcfg *plugin = (SCPluginIfcfg *)user_data;
	NMEtcnetConnection *connection = (NMEtcnetConnection *) val;
	g_signal_emit_by_name (plugin, NM_SYSTEM_CONFIG_INTERFACE_CONNECTION_ADDED, connection);
}

static void
connection_removed_cb(gpointer key, gpointer val, gpointer user_data)
{
	NMEtcnetConnection *connection = (NMEtcnetConnection *)val;
	nm_settings_connection_delete (NM_SETTINGS_CONNECTION (connection),
                                            ignore_cb,
                                            NULL);
}

static NMEtcnetDevice *
nm_etcnet_device_new (SCPluginIfcfg *plugin, const char *ifaces_path, const char *iface, const char *mac, NMDeviceType dev_type)
{
	NMEtcnetDevice *etcnet_device;

	g_return_val_if_fail (ifaces_path != NULL, NULL);
	g_return_val_if_fail (iface != NULL, NULL);
	g_return_val_if_fail (mac != NULL, NULL);

	if (dev_type != NM_DEVICE_TYPE_ETHERNET && dev_type != NM_DEVICE_TYPE_WIFI)
		return NULL;
	etcnet_device = (NMEtcnetDevice *)g_malloc0(sizeof(NMEtcnetDevice));
	if(!etcnet_device) {
		PLUGIN_WARNING ("not enough memory for NMEtcnetDevice struct.");
		return NULL;
	}

	etcnet_device->plugin = plugin;
	etcnet_device->iface_path = get_iface_path(ifaces_path, iface);
	etcnet_device->mac = g_strdup(mac);
	etcnet_device->unmanaged = FALSE;
	etcnet_device->dev_type = dev_type;
	etcnet_device->connections = g_hash_table_new_full(g_str_hash, is_profile_equal, g_free, g_object_unref);

	start_monitor(&etcnet_device->iface_monitor, etcnet_device->iface_path,
				G_CALLBACK (device_dir_changed), etcnet_device);

	return etcnet_device;
}

static void
nm_etcnet_device_free (gpointer data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *)data;
	g_return_if_fail (etcnet_device != NULL);

	g_free(etcnet_device->iface_path);
	g_free(etcnet_device->mac);

	g_hash_table_destroy(etcnet_device->connections);

	stop_monitor(&etcnet_device->iface_monitor);

	g_free(etcnet_device);
}

static void
reset_device_connections(NMEtcnetDevice *etcnet_device)
{
	g_hash_table_foreach (etcnet_device->connections, connection_removed_cb, NULL);
	g_hash_table_destroy(etcnet_device->connections);
	etcnet_device->connections = g_hash_table_new_full(g_str_hash, is_profile_equal, g_free, g_object_unref);
}

static void
update_device_config(NMEtcnetDevice *etcnet_device)
{
	gboolean old_unmanaged;

	g_return_if_fail (etcnet_device != NULL);

	old_unmanaged = etcnet_device->unmanaged;

	if(read_connections_for_one_iface(etcnet_device)) {
		if(!etcnet_device->unmanaged ) {
			g_hash_table_foreach (etcnet_device->connections, connection_added_cb, etcnet_device->plugin);
		}
		if (etcnet_device->unmanaged != old_unmanaged) {
			g_signal_emit_by_name (etcnet_device->plugin, NM_SYSTEM_CONFIG_INTERFACE_UNMANAGED_SPECS_CHANGED);
		}
	}
}

static gboolean
is_wireless (GUdevDevice *device)
{
	char phy80211_path[255];
	struct stat s;
	int fd;
	struct iwreq iwr;
	const char *ifname, *path;
	gboolean is_wifi = FALSE;

	ifname = g_udev_device_get_name (device);
	g_assert (ifname);

	fd = socket (PF_INET, SOCK_DGRAM, 0);
	strncpy (iwr.ifr_ifrn.ifrn_name, ifname, IFNAMSIZ);

	path = g_udev_device_get_sysfs_path (device);
	g_snprintf (phy80211_path, sizeof (phy80211_path), "%s/phy80211", path);

	if ((ioctl (fd, SIOCGIWNAME, &iwr) == 0)
			|| (stat (phy80211_path, &s) == 0 && (s.st_mode & S_IFDIR)))
		is_wifi = TRUE;

	close (fd);
	return is_wifi;
}

static NMDeviceType
get_dev_type(GUdevDevice *device)
{
	gint devtype;
	const char *sysfs_path;
	g_return_val_if_fail (device != NULL, NM_DEVICE_TYPE_UNKNOWN);

	devtype = g_udev_device_get_sysfs_attr_as_int (device, "type");
	sysfs_path = g_udev_device_get_sysfs_path (device);
	/* Check device type and ensure that it not virtual */
	if (devtype != 1 || g_str_has_prefix (sysfs_path, "/sys/devices/virtual"))
		return NM_DEVICE_TYPE_UNKNOWN;

	return is_wireless(device) ? NM_DEVICE_TYPE_WIFI : NM_DEVICE_TYPE_ETHERNET;
}

static NMEtcnetDevice *
read_device_config(SCPluginIfcfg *plugin, GUdevDevice *device)
{
	NMEtcnetDevice *etcnet_device;
	const char *iface;

	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);

	iface = g_udev_device_get_name (device);
	g_return_val_if_fail (iface != NULL, NULL);

	etcnet_device = g_hash_table_lookup(priv->devices, iface);
	if(!etcnet_device) {
		const char *mac;
		mac = g_udev_device_get_sysfs_attr (device, "address");
		etcnet_device = nm_etcnet_device_new(plugin, priv->ifaces_path, iface, mac, get_dev_type(device));
		if (etcnet_device == NULL)
			return NULL;
		g_hash_table_insert(priv->devices, g_strdup(iface), etcnet_device);
	} else {
		/*Device already in table*/
		/*FIXME: Update connections, not reset them */
		reset_device_connections(etcnet_device);
	}

	read_connections_for_one_iface(etcnet_device);

	return etcnet_device;
}

static void
read_connections (SCPluginIfcfg *plugin)
{
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);
	GList *list;
	GList *iter;

	list = g_udev_client_query_by_subsystem (priv->client, "net");
	for (iter = list; iter; iter = iter->next) {
		read_device_config (plugin, G_UDEV_DEVICE (iter->data));
		g_object_unref (G_UDEV_DEVICE (iter->data));
	}

	g_list_free (list);
}

static void
get_connections_cb (gpointer key, gpointer val, gpointer user_data)
{
	GSList **list = (GSList **) user_data;

	*list = g_slist_prepend (*list, val);
}

static void
get_devices_cb (gpointer key, gpointer val, gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) val;

	g_hash_table_foreach (etcnet_device->connections, get_connections_cb, user_data);
}

static GSList *
get_connections (NMSystemConfigInterface *config)
{
	SCPluginIfcfg *plugin = SC_PLUGIN_IFCFG (config);
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);
	GSList *list = NULL;

	if(!g_hash_table_size(priv->devices)) {
		read_connections (plugin);
	}

	g_hash_table_foreach (priv->devices, get_devices_cb, &list);

	return list;
}

static void
handle_uevent_cb (GUdevClient *client,
		const char *action,
		GUdevDevice *device,
		gpointer user_data)
{
	SCPluginIfcfg *plugin = SC_PLUGIN_IFCFG (user_data);
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);
	const char *subsys;
	NMEtcnetDevice *etcnet_device;

	g_return_if_fail (action != NULL);

	/* A bit paranoid */
	subsys = g_udev_device_get_subsystem (device);
	g_return_if_fail (subsys != NULL);
	g_return_if_fail (strcmp ("net", subsys) == 0);

	if (!strcmp ("add", action)) {
		etcnet_device = read_device_config (plugin, device);
		if (etcnet_device) {
			g_hash_table_foreach (etcnet_device->connections, connection_added_cb, plugin);
			g_signal_emit_by_name (plugin, NM_SYSTEM_CONFIG_INTERFACE_UNMANAGED_SPECS_CHANGED);
		}
	} else if (!strcmp ("remove", action)) {
		const char *iface;

		iface = g_udev_device_get_name (device);

		etcnet_device = (NMEtcnetDevice *)g_hash_table_lookup(priv->devices, iface);
		if(etcnet_device) {
			if(etcnet_device->unmanaged) {
				g_hash_table_remove (priv->devices, iface);
			} else {
				g_hash_table_foreach (etcnet_device->connections, connection_removed_cb, NULL);
				g_hash_table_remove (priv->devices, iface);
			}
		}
	}
}

static void
init (NMSystemConfigInterface *config)
{
	SCPluginIfcfg *plugin = SC_PLUGIN_IFCFG (config);
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);
	const char *subsys[2] = { "net", NULL };

	priv->ifaces_path = g_strdup(ETCNET_IFACES_PATH);
	if (!priv->ifaces_path)
		PLUGIN_WARNING("could not determine network ifaces path.");

	priv->client = g_udev_client_new (subsys);
	if (!priv->client) {
		PLUGIN_WARNING ("error initializing libgudev");
	} else
		g_signal_connect (priv->client, "uevent", G_CALLBACK (handle_uevent_cb), plugin);

	start_monitor (&priv->ifaces_monitor, priv->ifaces_path, G_CALLBACK(ifaces_dir_changed_cb), plugin);

	priv->hostname = get_hostname();
	start_monitor (&priv->network_file_monitor, NETWORK_FILE, G_CALLBACK(network_file_changed_cb), plugin);

	priv->current_profile = get_current_profile();
	PLUGIN_INFO ("Current network profile: %s",
			priv->current_profile && priv->current_profile[0] != '#' ?
			priv->current_profile : "(none)");
}

static void
sc_plugin_ifcfg_init (SCPluginIfcfg *plugin)
{
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);

	priv->devices = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, nm_etcnet_device_free);
}

static void
dispose (GObject *object)
{
	SCPluginIfcfg *plugin = SC_PLUGIN_IFCFG (object);
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (plugin);

	g_hash_table_destroy (priv->devices);

	g_free(priv->ifaces_path);
	stop_monitor (&priv->ifaces_monitor);
	stop_monitor (&priv->network_file_monitor);
	g_free (priv->hostname);

	g_free (priv->current_profile);

	G_OBJECT_CLASS (sc_plugin_ifcfg_parent_class)->dispose (object);
}

static void
finalize (GObject *object)
{
	G_OBJECT_CLASS (sc_plugin_ifcfg_parent_class)->finalize (object);
}

static void
get_property (GObject *object, guint prop_id,
			  GValue *value, GParamSpec *pspec)
{
	SCPluginIfcfgPrivate *priv = SC_PLUGIN_IFCFG_GET_PRIVATE (object);

	switch (prop_id) {
	case NM_SYSTEM_CONFIG_INTERFACE_PROP_NAME:
		g_value_set_string (value, IFCFG_PLUGIN_NAME);
		break;
	case NM_SYSTEM_CONFIG_INTERFACE_PROP_INFO:
		g_value_set_string (value, IFCFG_PLUGIN_INFO);
		break;
	case NM_SYSTEM_CONFIG_INTERFACE_PROP_CAPABILITIES:
		g_value_set_uint (value, NM_SYSTEM_CONFIG_INTERFACE_CAP_MODIFY_HOSTNAME);
		break;
	case NM_SYSTEM_CONFIG_INTERFACE_PROP_HOSTNAME:
		if (priv)
			g_value_set_string (value, priv->hostname);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
set_property (GObject *object, guint prop_id,
			  const GValue *value, GParamSpec *pspec)
{
	switch (prop_id) {
	case NM_SYSTEM_CONFIG_INTERFACE_PROP_HOSTNAME:
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
sc_plugin_ifcfg_class_init (SCPluginIfcfgClass *req_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (req_class);

	g_type_class_add_private (req_class, sizeof (SCPluginIfcfgPrivate));

	object_class->dispose = dispose;
	object_class->finalize = finalize;
	object_class->get_property = get_property;
	object_class->set_property = set_property;

	g_object_class_override_property (object_class,
							NM_SYSTEM_CONFIG_INTERFACE_PROP_NAME,
							NM_SYSTEM_CONFIG_INTERFACE_NAME);

	g_object_class_override_property (object_class,
							NM_SYSTEM_CONFIG_INTERFACE_PROP_INFO,
							NM_SYSTEM_CONFIG_INTERFACE_INFO);

	g_object_class_override_property (object_class,
							NM_SYSTEM_CONFIG_INTERFACE_PROP_CAPABILITIES,
							NM_SYSTEM_CONFIG_INTERFACE_CAPABILITIES);

	g_object_class_override_property (object_class,
							NM_SYSTEM_CONFIG_INTERFACE_PROP_HOSTNAME,
							NM_SYSTEM_CONFIG_INTERFACE_HOSTNAME);
}

static void
system_config_interface_init (NMSystemConfigInterface *system_config_interface_class)
{
	/* interface implementation */
	system_config_interface_class->get_connections = get_connections;
	system_config_interface_class->get_unmanaged_specs = get_unmanaged_specs;
	system_config_interface_class->init = init;
}

G_MODULE_EXPORT GObject *
nm_system_config_factory (const char *config_file)
{
	static SCPluginIfcfg *singleton = NULL;

	if (!singleton)
		singleton = SC_PLUGIN_IFCFG (g_object_new (SC_TYPE_PLUGIN_IFCFG, NULL));
	else
		g_object_ref (singleton);

	return G_OBJECT (singleton);
}
