/* NetworkManager system settings service
 *
 * Mikhail Efremov <sem@altlinux.org>
 *
 * Based on ifcfg-fedora and ifcfg-suse code
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2008 - 2011 Mikhail Efremov
 */

#include "common.h"
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <sys/inotify.h>
#include <errno.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <unistd.h>
/* For INT32_MAX */
#include <stdint.h>

#ifndef __user
#define __user
#endif
#include <linux/types.h>
#include <wireless.h>
#undef __user

#include <glib.h>

#include <nm-connection.h>
#include <NetworkManager.h>
#include <nm-setting-connection.h>
#include <nm-setting-ip4-config.h>
#include <nm-setting-wired.h>
#include <nm-setting-wireless.h>
#include <nm-setting-8021x.h>
#include <nm-utils.h>

#include "shvar.h"
#include "reader.h"

static gboolean ip6addr_from_str(const char *addr_str, struct in6_addr *out_addr);
static gboolean ip4addr_from_str(const char *addr_str, guint32 *out_addr);

static shvarFile *
svNewFile_iface_file(const char *iface_path, const char *file, const char *profile)
{
	char *filename_profile, *file_path;
	shvarFile *shvar_f = NULL;

	filename_profile = get_filename_with_profile(file, profile);
	file_path = g_build_filename(iface_path, filename_profile, NULL);
	shvar_f = svNewFile(file_path);
	if(!shvar_f && profile) {
		g_free(file_path);
		file_path = g_build_filename(iface_path, file, NULL);
		shvar_f = svNewFile(file_path);
	}

	g_free(file_path);
	g_free(filename_profile);

	return shvar_f;
}

typedef struct options_file_t
{
	shvarFile *options;
	shvarFile *default_options_eth;
	shvarFile *default_options;
}options_file_t;

static void
close_options_file(options_file_t *of)
{
	if(!of)
		return;

	if(of->options)
		svCloseFile(of->options);
	if(of->default_options_eth)
		svCloseFile(of->default_options_eth);
	if(of->default_options)
		svCloseFile(of->default_options);
	g_free(of);
}

static options_file_t *
open_options_file(const char *iface_path, const char *profile)
{
	options_file_t *of;
	char *ifaces_dir = NULL;
	char *def_opt_path = NULL;
	char *def_opt_eth_path = NULL;

	of = g_malloc0(sizeof(options_file_t));
	if(!of) {
		PLUGIN_WARNING("%s: Not enough memory for options_file_t", __func__);
		goto error;
	}
	of->options = svNewFile_iface_file(iface_path, "options", profile);

	ifaces_dir = g_path_get_dirname(iface_path);
	if(!ifaces_dir)
		goto error;

	def_opt_eth_path = g_build_filename(ifaces_dir, "default", "options-eth", NULL);
	if( !def_opt_eth_path)
		goto error;

	of->default_options_eth = svNewFile(def_opt_eth_path);

	def_opt_path = g_build_filename(ifaces_dir, "default", "options", NULL);
	if(!def_opt_path)
		goto error;

	of->default_options = svNewFile(def_opt_path);

	g_free(ifaces_dir);
	g_free(def_opt_eth_path);
	g_free(def_opt_path);

	return of;

error:
	g_free(ifaces_dir);
	g_free(def_opt_eth_path);
	g_free(def_opt_path);
	close_options_file(of);

	return NULL;
}

static char *
options_get_value(options_file_t *of, const char *key)
{
	char *value = NULL;

	g_return_val_if_fail(of != NULL, NULL);
	g_return_val_if_fail(key != NULL, NULL);

	if(of->options)
		value = svGetValue(of->options, key);
	if(!value && of->default_options_eth)
		value = svGetValue(of->default_options_eth, key);
	if(!value && of->default_options)
		value = svGetValue(of->default_options, key);

	return value;
}

static int
options_true_value(options_file_t *of, const char *key, int def)
{
	int ret = -1;

	g_return_val_if_fail(of != NULL, def);
	g_return_val_if_fail(key != NULL, def);

	if(of->options)
		ret = svTrueValue(of->options, key, -1);
	if(ret < 0 && of->default_options_eth)
		ret = svTrueValue(of->default_options_eth, key, -1);
	if(ret < 0)
		ret = of->default_options ? svTrueValue(of->default_options, key, def) : def;

	return ret;
}

static char *
read_file_contents(const char *iface_path, const char *file, const char *profile, GError **error)
{
	char *filename_profile, *file_fullname;
	char *contents = NULL;
	GError *err = NULL;

	g_return_val_if_fail(iface_path != NULL, NULL);
	g_return_val_if_fail(file != NULL, NULL);

	if(!error)
		error = &err;

	filename_profile = get_filename_with_profile(file, profile);
	file_fullname = g_build_filename(iface_path, filename_profile, NULL);

	if (!g_file_get_contents (file_fullname, &contents, NULL, error) && profile) {
		g_assert(error != NULL);

		if(g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
			g_free(file_fullname);
			g_clear_error(error);
			file_fullname = g_build_filename(iface_path, file, NULL);
			g_file_get_contents (file_fullname, &contents, NULL, error);
		}
	}

	if(error == &err && err) {
		g_error_free(err);
	}

	g_free(file_fullname);
	g_free(filename_profile);

	return contents;
}

static char **
read_file_lines(const char *iface_path, const char *file, const char *profile, GError **error)
{
	char **lines = NULL;
	char *contents = NULL;

	g_return_val_if_fail(iface_path != NULL, NULL);
	g_return_val_if_fail(file != NULL, NULL);

	contents = read_file_contents(iface_path, file, profile, error);
	if(contents) {
		lines = g_strsplit (contents, "\n", 0);
	}
	g_free(contents);

	return lines;
}

static gboolean
get_int (const char *str, long int *value)
{
	char *e;

	g_return_val_if_fail(str != NULL, FALSE);
	g_return_val_if_fail(value != NULL, FALSE);

	*value = strtol (str, &e, 0);
	if (*e != '\0')
		return FALSE;

	return TRUE;
}

static void
strip_line(char *line)
{
	char *p, *tmp;
	size_t len;
	int i, j;
	gboolean delim = FALSE;
	gboolean quoted = FALSE;

	g_return_if_fail (line != NULL);

	p = strchr (line, '#');
	if(p)
		*p = '\0';
	g_strstrip (line);

	/* Drop duplicate delimiters, but
	 * don't touch delimeters in quotes */
	len = strlen (line);

	tmp = g_malloc (len + 1);
	if (!tmp)
		return;

	for (i = 0, j = 0; i < len; i++) {
		if ((line[i] == ' ' || line[i] == '\t') && !quoted) {
			if (delim)
				continue;
			else
				delim = TRUE;
		} else
			delim = FALSE;

		if (line[i] == '\'' || line[i] == '\"')
			quoted = !quoted;

		tmp[j++] = line[i];
	}
	tmp[j] = '\0';

	if (j != len)
		g_strlcpy (line, tmp, len + 1);

	g_free (tmp);
}

typedef void (*strarray_callback_func) (const char *str, gpointer user_data);

/*
 * Call func for each line in str_array. Lines CAN BE modifyed.
 */
static void
str_array_foreach (char **str_array,
				strarray_callback_func func,
				gpointer user_data)
{
	char **pline;

	g_return_if_fail (str_array != NULL);
	g_return_if_fail (func != NULL);

	for(pline = str_array; *pline; ++pline) {
		strip_line(*pline);
		if(**pline != '\0')
			func (*pline, user_data);
	}
}

static gboolean
for_each_line (const char *iface_path, const char *file, const char *profile,
			strarray_callback_func func, gpointer user_data, GError **error)
{
	char **lines = NULL;

	g_return_val_if_fail (func != NULL, FALSE);

	lines = read_file_lines(iface_path, file, profile, error);
	if(!lines)
		return FALSE;

	str_array_foreach (lines, func, user_data);

	return TRUE;
}

static GHashTable *
read_supplicant_network(char ***plines)
{
	char **lines = *plines;
	GHashTable *network = NULL;
	char **line;
	char **values;

	g_return_val_if_fail(lines != NULL, NULL);

	network = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	if(!network) {
		PLUGIN_WARNING("%s: Not enough memory for GHashTable", __func__);
	}

	for(line = lines + 1; *line; ++line) {
		char *p;

		strip_line(*line);
		if(**line == '\0')
			continue;
		if(**line == '}')
			break;
		values = g_strsplit (*line, "=", 2);
		if( !values || !values[0] || !values[1])
			break;


		if(values[1][0] == '\"') {
			int len;
			p = values[1] + 1;
			len = strlen(values[1]);
			values[1][len - 1] = '\0';
		}
		else
			p = values[1];

		g_hash_table_insert(network, g_strdup(values[0]), g_strdup(p));
		g_strfreev(values);
	}
	*plines = line;

	if(**line != '}') {
		PLUGIN_WARNING("Not terminated network block!");
		g_hash_table_destroy(network);
		network = NULL;
	}

	return network;
}

GSList *
read_supplicant_config(const char *iface_path, const char *profile)
{
	GSList *wsconf_list = NULL;
	GHashTable *wsconf = NULL;
	char **lines = NULL;
	char **pline;

	lines = read_file_lines(iface_path, "wpa_supplicant.conf", profile, NULL);
	if(!lines) {
		goto out;
	}

	for(pline = lines; *pline; ++pline) {
		strip_line(*pline);
		if(**pline == '\0')
			continue;
		if(!strcmp("network={", *pline))
		{
			wsconf = read_supplicant_network(&pline);
			if(!wsconf) {
				PLUGIN_WARNING("Parsing wpa_supplicant.conf error");
				break;
			} else {
				wsconf_list = g_slist_prepend(wsconf_list, wsconf);
			}
		}
	}

out:

	g_strfreev(lines);
	return wsconf_list;
}

void
clear_supplicant_config(GSList *wsconfig_list)
{
	if(!wsconfig_list)
		return;

	g_slist_foreach(wsconfig_list, (GFunc) g_hash_table_destroy, NULL);
}

static NMSetting *
make_connection_setting (const char *iface_path,
                         const char *profile,
                         options_file_t *options,
                         const char *type,
                         const char *suggested,
                         gboolean *autoconnect)
{
	NMSettingConnection *s_con;
	char *iface_name = NULL;
	char *new_id = NULL, *uuid = NULL;
	char *str_for_uuid = NULL;

	g_return_val_if_fail(iface_path != NULL, NULL);
	g_return_val_if_fail(options != NULL, NULL);
	g_return_val_if_fail(type != NULL, NULL);
	g_return_val_if_fail(autoconnect != NULL, NULL);

	iface_name = get_iface_name (iface_path);
	if (!iface_name)
		return NULL;

	s_con = NM_SETTING_CONNECTION (nm_setting_connection_new ());

	if (suggested) {
		/* For cosmetic reasons, if the suggested name is the same as
		 * the ifcfg files name, don't use it.
		 */
		if (strcmp (iface_name, suggested)) {
			new_id = g_strdup_printf ("%s %s (%s)", (profile != NULL ? profile : "System"),
									suggested, iface_name);
		}
	}

	if (!new_id) {
		new_id = g_strdup_printf ("%s %s", (profile != NULL ? profile : "System"), iface_name);
	}

	g_object_set (s_con,
			    NM_SETTING_CONNECTION_ID, new_id,
			    NM_SETTING_CONNECTION_TYPE, type,
			    NM_SETTING_CONNECTION_READ_ONLY, TRUE,
			    NULL);

	/* In case of suggested == NULL it is just like g_strdup().
	 * Otherwise suggested will be appended.
	 */
	str_for_uuid = g_strconcat (options->options ? options->options->fileName : iface_path,
			                        suggested, NULL);

	uuid = nm_utils_uuid_generate_from_string (str_for_uuid);
	g_object_set (s_con, NM_SETTING_CONNECTION_UUID, uuid, NULL);

	g_free (str_for_uuid);
	g_free (uuid);
	g_free (new_id);

	/* Be somewhat conservative about autoconnect,
	 * but set autoconnection property always to FALSE
	 * (default value for autoconnect is TRUE).
	 * The proper value for the current profile
	 * will be setted later. */
	*autoconnect = options_true_value (options, "ONBOOT", FALSE) > 0;
	g_object_set (s_con, NM_SETTING_CONNECTION_AUTOCONNECT, FALSE, NULL);

	g_free (iface_name);
	return (NMSetting *) s_con;
}

typedef struct iplink_info_t
{
	char *mac;
	char *mtu;
} iplink_info_t;

static iplink_info_t *
iplink_info_alloc(void)
{
	return g_malloc0(sizeof(iplink_info_t));
}

static void
iplink_info_free(iplink_info_t *info)
{
	if(!info)
		return;

	g_free(info->mac);
	g_free(info->mtu);
	g_free(info);
}

static iplink_info_t *
get_iplink_info(const char *iface_path, const char *profile, GError **error)
{
	iplink_info_t *info = NULL;
	char **lines;
	char **pline;

	info = iplink_info_alloc();

	lines = read_file_lines(iface_path, "iplink", profile, error);
	if(!lines) {
		if(g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
			g_clear_error(error);
		}

		return info;
	}

	for(pline = lines; *pline; ++pline) {
		char **values;
		strip_line(*pline);
		if(**pline == '\0')
			continue;
		values = g_strsplit_set (*pline, " \t", 0);
		if(!values)
			continue;
		if(values[0] && values[1]) {
			g_strstrip(values[0]);
			g_strstrip(values[1]);
			if(!g_ascii_strcasecmp (values[0], "address")) {
				info->mac = g_strdup(values[1]);
			} else if(!g_ascii_strcasecmp (values[0], "mtu")) {
				info->mtu = g_strdup(values[1]);
			}
		}
		g_strfreev(values);
	}
	g_strfreev(lines);

	return info;
}

/***********************************************/
/* Common IP functions and types               */
/***********************************************/
typedef struct add_routes_info_t
{
	GSList **p_routes_list;
	GSList *addr_list;
} add_routes_info_t;

/***********************************************/
/* IPv4 functions                              */
/***********************************************/
static gboolean
ip4addr_from_str(const char *addr_str, guint32 *out_addr)
{
	struct in_addr ip4_addr;

	g_return_val_if_fail(addr_str != NULL, FALSE);
	g_return_val_if_fail(out_addr != NULL, FALSE);

	if (inet_pton (AF_INET, addr_str, &ip4_addr) > 0) {
		*out_addr = ip4_addr.s_addr;
		return TRUE;
	}

	return FALSE;
}

static void
set_ipv4dns_config_cb (const char *line, gpointer user_data)
{
	NMSettingIP4Config *s_ip4 = (NMSettingIP4Config *) user_data;
	char **str_array = NULL;
	guint array_size;

	str_array = g_strsplit_set (line, " \t", 0);
	if (!str_array)
		return;

	array_size = g_strv_length (str_array);
	/* Should be at least 2 tokens */
	if (array_size < 2)
		goto done;

	if (!strcmp (str_array[0], "search") ||
			!strcmp (str_array[0], "domain")) {
		char **item;
		/* Skip first element in array. */
		for (item = str_array + 1; *item; item++)
			nm_setting_ip4_config_add_dns_search(s_ip4, *item);
	} else if (array_size == 2 &&
			!strcmp (str_array[0], "nameserver")) {
		guint32 ipv4dns;
		struct in6_addr ipv6dns;

		/* Second element is IP address of the nameserver. */
		if (ip4addr_from_str(str_array[1], &ipv4dns))
			nm_setting_ip4_config_add_dns(s_ip4, ipv4dns);
		else if (ip6addr_from_str(str_array[1], &ipv6dns)) {
			/* Just skip IPv6 addresses */
			goto done;
		} else
			PLUGIN_WARNING ("Invalid IPv4 DNS server address '%s'", str_array[1]);
	}
done:
	if (str_array)
		g_strfreev (str_array);
}

/* Get IPv4 address and prefix from string */
static gboolean
split_ip4addr_with_prefix_str (const char *str, guint32 *addr, guint32 *prefix)
{
	gchar **str_array = NULL;
	gboolean ret = FALSE;
	long int tmp;

	g_return_val_if_fail (str != NULL, FALSE);
	g_return_val_if_fail (addr != NULL, FALSE);
	g_return_val_if_fail (prefix != NULL, FALSE);

	str_array = g_strsplit(str, "/", 0);

	if(str_array && str_array[0] && str_array[1] &&
			ip4addr_from_str(str_array[0], addr) &&
			get_int(str_array[1], &tmp) &&
			tmp >= 0 && tmp <= 32) {
		*prefix = (guint32) tmp;
		ret = TRUE;
	}

	g_strfreev(str_array);

	return ret;
}

static gboolean
set_ip4addr_and_prefix (NMIP4Address *nm_ip4_addr, const char *addr_str,
						GError **error)
{
	guint32 addr, prefix;

	g_return_val_if_fail (nm_ip4_addr != NULL, FALSE);
	g_return_val_if_fail (addr_str != NULL, FALSE);

	if (!split_ip4addr_with_prefix_str (addr_str, &addr, &prefix)) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Invalid IPv4 address and prefix string: '%s'", addr_str);
		return FALSE;
	}

	nm_ip4_address_set_address (nm_ip4_addr, addr);
	nm_ip4_address_set_prefix (nm_ip4_addr, prefix);

	return TRUE;
}

static void
set_ip4gw_cb (gpointer data, gpointer user_data)
{
	NMIP4Address *addr = (NMIP4Address *) data;
	guint32 gw = (guint32) GPOINTER_TO_UINT(user_data);

	g_return_if_fail (addr != NULL);

	nm_ip4_address_set_gateway (addr, gw);
}

static gboolean
set_ip4route (NMIP4Route **route, GSList *addr_list, const char *line, GError **error)
{
	gchar **str_array = NULL;
	gboolean ret = FALSE;
	int i, gw_index = -1;
	gboolean def_gw = FALSE;
	guint32 gw_addr, metric = 0;

	g_return_val_if_fail (route != NULL, FALSE);
	g_return_val_if_fail (addr_list != NULL, FALSE);
	g_return_val_if_fail (line != NULL, FALSE);

	str_array = g_strsplit_set (line, " \t", 0);
	if(!str_array || !str_array[0] || !str_array[1]) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Invalid IPv4 route string: '%s'", line);
		goto done;
	}

	/* Is it default gateway? */
	if (g_ascii_strcasecmp (str_array[0], "default") == 0)
		def_gw = TRUE;

	/* 'via' can not be first */
	for (i = 1; str_array[i] != NULL; ++i) {
		/* search word 'via' in string, next - IP address */
		if (!g_ascii_strcasecmp (str_array[i], "via")) {
			if (str_array[++i] == NULL)
				break;
			gw_index = i;
			continue;
		}
		/* find metric value in string */
		if (!g_ascii_strcasecmp (str_array[i], "metric")) {
			long int tmp;

			if (str_array[++i] == NULL)
				break;
			if (get_int (str_array[i], &tmp) && tmp >= 0 && tmp <= INT32_MAX)
				metric = (guint32) tmp;
			else
				PLUGIN_WARNING ("Invalid metric: %s", str_array[i]);
		}
	}

	if (gw_index == -1) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
					"Invalid IPv4 route string: '%s'", line);
		goto done;
	}

	if (!ip4addr_from_str(str_array[gw_index], &gw_addr)) {
			g_set_error (error, ifcfg_plugin_error_quark (), 0,
			             "Invalid IPv4 gateway '%s'", str_array[gw_index]);
			goto done;
	}

	if (def_gw) {
		/* Set gateway for all addresses */
		g_slist_foreach (addr_list, set_ip4gw_cb, GUINT_TO_POINTER(gw_addr));
	} else {
		/* Add route */
		guint32 addr, prefix;

		if (split_ip4addr_with_prefix_str (str_array[0], &addr, &prefix)) {
			*route = nm_ip4_route_new ();
			nm_ip4_route_set_dest (*route, addr);
			nm_ip4_route_set_prefix (*route, prefix);
			nm_ip4_route_set_next_hop (*route, gw_addr);
			nm_ip4_route_set_metric (*route, metric);
		} else {
			g_set_error (error, ifcfg_plugin_error_quark (), 0,
			             "Invalid IPv4 route: '%s'", line);
			goto done;
		}
	}

	ret = TRUE;
done:
	g_strfreev(str_array);

	return ret;
}

/*
 * Fill list with IPv4 addresses.
 */
static void
add_ip4addr_cb (const char *line, gpointer user_data)
{
	NMIP4Address *addr = NULL;
	GError *error = NULL;
	GSList **list = (GSList **) user_data;

	g_return_if_fail (list != NULL);

	addr = nm_ip4_address_new ();

	if (set_ip4addr_and_prefix(addr, line, &error))
		*list = g_slist_prepend (*list, addr);
	else {
		PLUGIN_ERROR("add IPv4 address", error);
		g_error_free (error);
		nm_ip4_address_unref (addr);
	}
}

/*
 * Fill list with IPv4 routes and set default gateway
 * for addresses.
 */
static void
add_ip4route_cb (const char *line, gpointer user_data)
{
	NMIP4Route *route = NULL;
	GError *error = NULL;
	add_routes_info_t *info = (add_routes_info_t *) user_data;

	g_return_if_fail (info != NULL);
	g_return_if_fail (info->p_routes_list != NULL);
	g_return_if_fail (info->addr_list != NULL);

	if (set_ip4route (&route, info->addr_list, line, &error)) {
		/* Add route to list if it is not default gateway.
		 * Otherwise it was already added to addresses. */
		if (route)
			*info->p_routes_list = g_slist_prepend (*info->p_routes_list, route);
	}else {
		PLUGIN_ERROR("add IPv4 route", error);
		g_error_free (error);
	}
}

static void
ip4_settings_add_address_cb (gpointer data, gpointer user_data)
{
	NMIP4Address *addr = (NMIP4Address *) data;
	NMSettingIP4Config *s_ip4 = (NMSettingIP4Config *) user_data;

	g_return_if_fail (addr != NULL);
	g_return_if_fail (s_ip4 != NULL);

	nm_setting_ip4_config_add_address (s_ip4, addr);
}

static void
ip4_settings_add_route_cb (gpointer data, gpointer user_data)
{
	NMIP4Route *route = (NMIP4Route *) data;
	NMSettingIP4Config *s_ip4 = (NMSettingIP4Config *) user_data;

	g_return_if_fail (route != NULL);
	g_return_if_fail (s_ip4 != NULL);

	nm_setting_ip4_config_add_route (s_ip4, route);
}

static NMSetting *
make_ip4_setting (options_file_t *options, const char *iface_path, const char *profile, GError **error)
{
	NMSettingIP4Config *s_ip4 = NULL;
	char *value = NULL;
	char *method = NM_SETTING_IP4_CONFIG_METHOD_AUTO;
	GSList *ip4addr_list = NULL;
	GSList *ip4routes_list = NULL;
	add_routes_info_t info;

	g_return_val_if_fail(options != NULL, NULL);
	g_return_val_if_fail(iface_path != NULL, NULL);

	value = options_get_value (options, "BOOTPROTO");
	if (value) {
		if( !g_ascii_strncasecmp (value, "static", 6) ||
				/* Yes, it is really static configuration for IPv4
				 * according with that brain-damaged logic in etcnet */
				!g_ascii_strncasecmp (value, "dhcp6", 5))
			method = NM_SETTING_IP4_CONFIG_METHOD_MANUAL;
		else if ( !g_ascii_strncasecmp (value, "ipv4ll", 6) )
			method = NM_SETTING_IP4_CONFIG_METHOD_LINK_LOCAL;
	}
	g_free (value);

	if(strcmp(method, NM_SETTING_IP4_CONFIG_METHOD_MANUAL) != 0)
		goto done;

	if(for_each_line (iface_path, "ipv4address", profile,
					add_ip4addr_cb, &ip4addr_list, error) && ip4addr_list) {
		/* Reverse the list for matching with order in config file. */
		ip4addr_list = g_slist_reverse (ip4addr_list);
	} else
		goto error;

	info.p_routes_list = &ip4routes_list;
	info.addr_list = ip4addr_list;
	if (for_each_line (iface_path, "ipv4route", profile,
					add_ip4route_cb, &info, error)) {
		/* Reverse the list for matching with order in config file. */
		if (ip4routes_list)
			ip4routes_list = g_slist_reverse (ip4routes_list);
	} else if(g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
		/* Don't care if ipv4route file does not exist */
		g_clear_error(error);
	} else
		goto error;

done:
	s_ip4 = (NMSettingIP4Config *) nm_setting_ip4_config_new ();
	if(!s_ip4) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Not enough memory for NMSettingIP4Config");
		goto error;
	}

	g_object_set (s_ip4, NM_SETTING_IP4_CONFIG_METHOD, method, NULL);

	if (ip4addr_list) {
		g_slist_foreach (ip4addr_list, ip4_settings_add_address_cb, s_ip4);
		g_slist_free_full (ip4addr_list, (GDestroyNotify) nm_ip4_address_unref);
		ip4addr_list = NULL;
	}

	if (ip4routes_list) {
		g_slist_foreach (ip4routes_list, ip4_settings_add_route_cb, s_ip4);
		g_slist_free_full (ip4routes_list, (GDestroyNotify) nm_ip4_route_unref);
		ip4routes_list = NULL;
	}

	/* Read resolv.conf */
	if (!for_each_line (iface_path, "resolv.conf", profile,
					set_ipv4dns_config_cb, s_ip4, error) &&
			g_error_matches (*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
		/* Don't care if resolv.conf file does not exist */
		g_clear_error(error);
	}

	return NM_SETTING (s_ip4);

error:
	if(ip4addr_list)
		g_slist_free_full (ip4addr_list, (GDestroyNotify) nm_ip4_address_unref);
	if(ip4routes_list)
		g_slist_free_full (ip4routes_list, (GDestroyNotify) nm_ip4_route_unref);
	if (s_ip4)
		g_object_unref (s_ip4);
	return NULL;
}

/***********************************************/
/* IPv6 functions                              */
/***********************************************/
static gboolean
ip6addr_from_str(const char *addr_str, struct in6_addr *out_addr)
{
	struct in6_addr ip6_addr;

	g_return_val_if_fail(addr_str != NULL, FALSE);
	g_return_val_if_fail(out_addr != NULL, FALSE);

	if (inet_pton (AF_INET6, addr_str, &ip6_addr) > 0) {
		*out_addr = ip6_addr;
		return TRUE;
	}

	return FALSE;
}

static void
set_ipv6dns_config_cb (const char *line, gpointer user_data)
{
	NMSettingIP6Config *s_ip6 = (NMSettingIP6Config *) user_data;
	char **str_array = NULL;
	guint array_size;

	str_array = g_strsplit_set (line, " \t", 0);
	if (!str_array)
		return;

	array_size = g_strv_length (str_array);
	/* Should be at least 2 tokens */
	if (array_size < 2)
		goto done;

	if (!strcmp (str_array[0], "search") ||
			!strcmp (str_array[0], "domain")) {
		char **item;
		/* Skip first element in array. */
		for (item = str_array + 1; *item; item++)
			nm_setting_ip6_config_add_dns_search(s_ip6, *item);
	} else if (array_size == 2 &&
			!strcmp (str_array[0], "nameserver")) {
		guint32 ipv4dns;
		struct in6_addr ipv6dns;

		/* Second element is IP address of the nameserver. */
		if (ip6addr_from_str(str_array[1], &ipv6dns))
			nm_setting_ip6_config_add_dns(s_ip6, &ipv6dns);
		else if (ip4addr_from_str(str_array[1], &ipv4dns)) {
			/* Just skip IPv4 addresses */
			goto done;
		} else
			PLUGIN_WARNING ("Invalid IPv6 DNS server address '%s'", str_array[1]);
	}
done:
	if (str_array)
		g_strfreev (str_array);
}

/* Get IPv6 address and prefix from string */
static gboolean
split_ip6addr_with_prefix_str (const char *str, struct in6_addr *addr, guint32 *prefix)
{
	gchar **str_array = NULL;
	gboolean ret = FALSE;
	long int tmp;

	g_return_val_if_fail (str != NULL, FALSE);
	g_return_val_if_fail (addr != NULL, FALSE);
	g_return_val_if_fail (prefix != NULL, FALSE);

	str_array = g_strsplit(str, "/", 0);

	if(str_array && str_array[0] && str_array[1] &&
			ip6addr_from_str(str_array[0], addr) &&
			get_int(str_array[1], &tmp) &&
			tmp >= 0 && tmp <= 128) {
		*prefix = (guint32) tmp;
		ret = TRUE;
	}

	g_strfreev(str_array);

	return ret;
}

static gboolean
set_ip6addr_and_prefix (NMIP6Address *nm_ip6_addr, const char *addr_str,
						GError **error)
{
	guint32 prefix;
	struct in6_addr addr;

	g_return_val_if_fail (nm_ip6_addr != NULL, FALSE);
	g_return_val_if_fail (addr_str != NULL, FALSE);

	if (!split_ip6addr_with_prefix_str (addr_str, &addr, &prefix)) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Invalid IPv6 address and prefix string: '%s'", addr_str);
		return FALSE;
	}

	nm_ip6_address_set_address (nm_ip6_addr, &addr);
	nm_ip6_address_set_prefix (nm_ip6_addr, prefix);

	return TRUE;
}

static void
set_ip6gw_cb (gpointer data, gpointer user_data)
{
	NMIP6Address *addr = (NMIP6Address *) data;
	struct in6_addr *gw = (struct in6_addr *) user_data;

	g_return_if_fail (addr != NULL);

	nm_ip6_address_set_gateway (addr, gw);
}

static gboolean
set_ip6route (NMIP6Route **route, GSList *addr_list, const char *line, GError **error)
{
	gchar **str_array = NULL;
	gboolean ret = FALSE;
	int i, gw_index = -1;
	gboolean def_gw = FALSE;
	guint32 metric = 0;
	struct in6_addr gw_addr;

	g_return_val_if_fail (route != NULL, FALSE);
	g_return_val_if_fail (addr_list != NULL, FALSE);
	g_return_val_if_fail (line != NULL, FALSE);

	str_array = g_strsplit_set (line, " \t", 0);
	if(!str_array || !str_array[0] || !str_array[1]) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Invalid IPv6 route string: '%s'", line);
		goto done;
	}

	/* Is it default gateway? */
	if (g_ascii_strcasecmp (str_array[0], "default") == 0)
		def_gw = TRUE;

	/* 'via' can not be first */
	for (i = 1; str_array[i] != NULL; ++i) {
		/* search word 'via' in string, next - IP address */
		if (!g_ascii_strcasecmp (str_array[i], "via")) {
			if (str_array[++i] == NULL)
				break;
			gw_index = i;
			continue;
		}
		/* find metric value in string */
		if (!g_ascii_strcasecmp (str_array[i], "metric")) {
			long int tmp;

			if (str_array[++i] == NULL)
				break;
			if (get_int (str_array[i], &tmp) && tmp >= 0 && tmp <= INT32_MAX)
				metric = (guint32) tmp;
			else
				PLUGIN_WARNING ("Invalid metric: %s", str_array[i]);
		}
	}

	if (gw_index == -1) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
					"Invalid IPv6 route string: '%s'", line);
		goto done;
	}

	if (!ip6addr_from_str(str_array[gw_index], &gw_addr)) {
			g_set_error (error, ifcfg_plugin_error_quark (), 0,
			             "Invalid IPv6 gateway '%s'", str_array[gw_index]);
			goto done;
	}

	if (def_gw) {
		/* Set gateway for all addresses */
		g_slist_foreach (addr_list, set_ip6gw_cb, &gw_addr);
	} else {
		/* Add route */
		guint32  prefix;
		struct in6_addr addr;

		if (split_ip6addr_with_prefix_str (str_array[0], &addr, &prefix)) {
			*route = nm_ip6_route_new ();
			nm_ip6_route_set_dest (*route, &addr);
			nm_ip6_route_set_prefix (*route, prefix);
			nm_ip6_route_set_next_hop (*route, &gw_addr);
			nm_ip6_route_set_metric (*route, metric);
		} else {
			g_set_error (error, ifcfg_plugin_error_quark (), 0,
			             "Invalid IPv6 route: '%s'", line);
			goto done;
		}
	}

	ret = TRUE;
done:
	g_strfreev(str_array);

	return ret;
}

/*
 * Fill list with IPv6 addresses.
 */
static void
add_ip6addr_cb (const char *line, gpointer user_data)
{
	NMIP6Address *addr = NULL;
	GError *error = NULL;
	GSList **list = (GSList **) user_data;

	g_return_if_fail (list != NULL);

	addr = nm_ip6_address_new ();

	if (set_ip6addr_and_prefix(addr, line, &error))
		*list = g_slist_prepend (*list, addr);
	else {
		PLUGIN_ERROR("add IPv6 address", error);
		g_error_free (error);
		nm_ip6_address_unref (addr);
	}
}

/*
 * Fill list with IPv4 routes and set default gateway
 * for addresses.
 */
static void
add_ip6route_cb (const char *line, gpointer user_data)
{
	NMIP6Route *route = NULL;
	GError *error = NULL;
	add_routes_info_t *info = (add_routes_info_t *) user_data;

	g_return_if_fail (info != NULL);
	g_return_if_fail (info->p_routes_list != NULL);
	g_return_if_fail (info->addr_list != NULL);

	if (set_ip6route (&route, info->addr_list, line, &error)) {
		/* Add route to list if it is not default gateway.
		 * Otherwise it was already added to addresses. */
		if (route)
			*info->p_routes_list = g_slist_prepend (*info->p_routes_list, route);
	}else {
		PLUGIN_ERROR("add IPv6 route", error);
		g_error_free (error);
	}
}

static void
ip6_settings_add_address_cb (gpointer data, gpointer user_data)
{
	NMIP6Address *addr = (NMIP6Address *) data;
	NMSettingIP6Config *s_ip6 = (NMSettingIP6Config *) user_data;

	g_return_if_fail (addr != NULL);
	g_return_if_fail (s_ip6 != NULL);

	nm_setting_ip6_config_add_address (s_ip6, addr);
}

static void
ip6_settings_add_route_cb (gpointer data, gpointer user_data)
{
	NMIP6Route *route = (NMIP6Route *) data;
	NMSettingIP6Config *s_ip6 = (NMSettingIP6Config *) user_data;

	g_return_if_fail (route != NULL);
	g_return_if_fail (s_ip6 != NULL);

	nm_setting_ip6_config_add_route (s_ip6, route);
}

static NMSetting *
make_ip6_setting (options_file_t *options, const char *iface_path, const char *profile, GError **error)
{
	NMSettingIP6Config *s_ip6 = NULL;
	char *value = NULL;
	char *method = NM_SETTING_IP6_CONFIG_METHOD_AUTO;
	GSList *ip6addr_list = NULL;
	GSList *ip6routes_list = NULL;
	add_routes_info_t info;

	g_return_val_if_fail(options != NULL, NULL);
	g_return_val_if_fail(iface_path != NULL, NULL);

	value = options_get_value (options, "BOOTPROTO");
	if (value) {
		if( !g_ascii_strncasecmp (value, "static", 6))
			/* Assume static configuration, but it can be RA
			 * if there is no a configuration.
			 * Check it later. */
			method = NM_SETTING_IP6_CONFIG_METHOD_MANUAL;
	}
	g_free (value);

	if(strcmp(method, NM_SETTING_IP4_CONFIG_METHOD_MANUAL) != 0)
		goto done;

	if(for_each_line (iface_path, "ipv6address", profile,
					add_ip6addr_cb, &ip6addr_list, error) && ip6addr_list) {
		/* Reverse the list for matching with order in config file. */
		ip6addr_list = g_slist_reverse (ip6addr_list);
	} else
		goto error;

	info.p_routes_list = &ip6routes_list;
	info.addr_list = ip6addr_list;
	if (for_each_line (iface_path, "ipv6route", profile,
					add_ip6route_cb, &info, error)) {
		/* Reverse the list for matching with order in config file. */
		if (ip6routes_list)
			ip6routes_list = g_slist_reverse (ip6routes_list);
	} else if(g_error_matches(*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
		/* Don't care if ipv6route file does not exist */
		g_clear_error(error);
	} else
		goto error;

	/* If there is no configuration assime it is RA.
	 * There is no such method like RA-only in NM, so
	 * set it to NM_SETTING_IP6_CONFIG_METHOD_AUTO. */
	if (!ip6routes_list && !ip6addr_list)
		method = NM_SETTING_IP6_CONFIG_METHOD_AUTO;

done:
	s_ip6 = (NMSettingIP6Config *) nm_setting_ip6_config_new ();
	if(!s_ip6) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Not enough memory for NMSettingIP6Config");
		goto error;
	}

	g_object_set (s_ip6, NM_SETTING_IP6_CONFIG_METHOD, method, NULL);

	if (ip6addr_list) {
		g_slist_foreach (ip6addr_list, ip6_settings_add_address_cb, s_ip6);
		g_slist_free_full (ip6addr_list, (GDestroyNotify) nm_ip6_address_unref);
		ip6addr_list = NULL;
	}

	if (ip6routes_list) {
		g_slist_foreach (ip6routes_list, ip6_settings_add_route_cb, s_ip6);
		g_slist_free_full (ip6routes_list, (GDestroyNotify) nm_ip6_route_unref);
		ip6routes_list = NULL;
	}

	/* Read resolv.conf */
	if (!for_each_line (iface_path, "resolv.conf", profile,
					set_ipv6dns_config_cb, s_ip6, error) &&
			g_error_matches (*error, G_FILE_ERROR, G_FILE_ERROR_NOENT)) {
		/* Don't care if resolv.conf file does not exist */
		g_clear_error(error);
	}

	return NM_SETTING (s_ip6);

error:
	if(ip6addr_list)
		g_slist_free_full (ip6addr_list, (GDestroyNotify) nm_ip6_address_unref);
	if(ip6routes_list)
		g_slist_free_full (ip6routes_list, (GDestroyNotify) nm_ip6_route_unref);
	if (s_ip6)
		g_object_unref (s_ip6);
	return NULL;
}

/***********************************************/
static GByteArray *
get_new_mac (iplink_info_t *info)
{
	GByteArray *mac = NULL;

	g_return_val_if_fail(info != NULL, NULL);

	if(!info->mac)
		return NULL;

	mac = get_ether_addr_array(info->mac);
	if (!mac) {
		PLUGIN_WARNING("The MAC address '%s' was invalid.", info->mac);
	}

	return mac;
}

static gboolean
get_mtu(iplink_info_t *info, int *mtu)
{
	long int tmp = -1;

	g_return_val_if_fail(info != NULL, FALSE);
	g_return_val_if_fail(mtu != NULL, FALSE);

	if (info->mtu && get_int (info->mtu, &tmp)) {
		if (tmp >= 0 && tmp < 65536) {
			*mtu = tmp;
			return TRUE;
		} else
			PLUGIN_WARNING("Invalid MTU '%s'", info->mtu);
	}

	return FALSE;
}

static NMSetting *
make_wired_setting (const char *iface_path, const char *profile, GError **error)
{
	NMSettingWired *s_wired;
	int mtu;
	GByteArray *mac = NULL;
	iplink_info_t *info;

	s_wired = NM_SETTING_WIRED (nm_setting_wired_new ());

	info = get_iplink_info(iface_path, profile, error);
	if(info) {
		if(get_mtu(info, &mtu))
			g_object_set (s_wired, NM_SETTING_WIRED_MTU, mtu, NULL);

		mac = get_new_mac (info);
		if(mac) {
			g_object_set (s_wired, NM_SETTING_WIRED_CLONED_MAC_ADDRESS, mac, NULL);
			g_byte_array_free (mac, TRUE);
		}
	}

	iplink_info_free(info);
	return (NMSetting *) s_wired;
}

static NMConnection *
read_wired_connection (const char *iface_path, const char *profile, options_file_t *options,
					gboolean *autoconnect, GError **error)
{
	NMConnection *connection = NULL;
	NMSetting *con_setting = NULL;
	NMSetting *wired_setting = NULL;

	g_return_val_if_fail (iface_path != NULL, NULL);
	g_return_val_if_fail (options != NULL, NULL);

	connection = nm_connection_new ();
	if (!connection) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Failed to allocate new connection for %s.", iface_path);
		return NULL;
	}

	con_setting = make_connection_setting (iface_path, profile, options,
			NM_SETTING_WIRED_SETTING_NAME, NULL, autoconnect);
	if (!con_setting) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Failed to create connection setting.");
		goto error;
	}
	nm_connection_add_setting (connection, con_setting);

	wired_setting = make_wired_setting (iface_path, profile, error);
	if (!wired_setting)
		goto error;

	nm_connection_add_setting (connection, wired_setting);

	if (!nm_connection_verify (connection, error)) {
		goto error;
	}

	return connection;

error:

	g_object_unref (connection);
	if (con_setting)
		g_object_unref (con_setting);
	if (wired_setting)
		g_object_unref (wired_setting);
	return NULL;
}

static char *
get_supplicant_config_value (GHashTable *wsconf, const char *key, gboolean to_lower)
{
	char *value;
	g_return_val_if_fail(wsconf != NULL, NULL);
	g_return_val_if_fail(key != NULL, NULL);

	value = g_hash_table_lookup(wsconf, key);
	if(!value)
		return NULL;
	return to_lower ? g_ascii_strdown(value, -1) : g_strdup(value);
}

static GSList *
get_supplicant_config_value_list (GHashTable *wsconf, const char *key, gboolean to_lower)
{
	char *values;
	gchar **str_array;
	int i;
	GSList *list = NULL;

	values = get_supplicant_config_value (wsconf, key, to_lower);

	if(!values)
		return NULL;
	str_array = g_strsplit_set (values, " \t", 0);
	if(str_array) {
		for(i = 0; str_array[i] != NULL; ++i) {
			list = g_slist_prepend (list, g_strdup(str_array[i]));
		}

		g_strfreev (str_array);
	}

	return list;
}

enum wireless_mode
{
	WL_MODE_INFRASTRUCTURE,
	WL_MODE_ADHOC,
	WL_MODE_AP,
	WL_MODE_UNKNOWN
};

static enum wireless_mode
get_wireless_mode (GHashTable *wsconf)
{
	char *mode_str;
	enum wireless_mode mode = WL_MODE_INFRASTRUCTURE;

	g_return_val_if_fail(wsconf != NULL, WL_MODE_UNKNOWN);

	mode_str = get_supplicant_config_value (wsconf, "mode", FALSE);

	if(mode_str) {
		if ( !g_ascii_strcasecmp (mode_str, "0") )
			mode = WL_MODE_INFRASTRUCTURE;
		else if ( !g_ascii_strcasecmp (mode_str, "1") )
			mode = WL_MODE_ADHOC;
		else if ( !g_ascii_strcasecmp (mode_str, "2") )
			mode = WL_MODE_AP;
		else
			mode = WL_MODE_UNKNOWN;
	}

	g_free(mode_str);

	return mode;
}

static char *
get_ssid (GHashTable *wsconf)
{
	return get_supplicant_config_value (wsconf, "ssid", FALSE);
}

static GByteArray *
get_bssid (GHashTable *wsconf)
{
	char *bssid_str = get_supplicant_config_value (wsconf, "bssid", FALSE);

	return bssid_str ? get_ether_addr_array(bssid_str) : NULL;
}

static char *
get_key_mgmt(GHashTable *wsconf)
{
	char *key_mgmt;

	key_mgmt = get_supplicant_config_value (wsconf, "key_mgmt", TRUE);
	if( !key_mgmt )
		key_mgmt = g_strdup ("none");

	return key_mgmt;
}

static guint32
get_wep_tx_keyidx(GHashTable *wsconf)
{
	long int wep_tx_keyidx = 0;
	char *wep_tx_keyidx_str;

	wep_tx_keyidx_str = get_supplicant_config_value (wsconf, "wep_tx_keyidx", TRUE);
	if(wep_tx_keyidx_str) {
		if (!get_int(wep_tx_keyidx_str, &wep_tx_keyidx) ||
				wep_tx_keyidx < 0 || wep_tx_keyidx > 3) {
			wep_tx_keyidx = 0;
		}
	}

	return (guint32)wep_tx_keyidx;
}

static char *
get_auth_alg(GHashTable *wsconf)
{
	return get_supplicant_config_value (wsconf, "auth_alg", TRUE);
}

static GSList *
get_proto(GHashTable *wsconf)
{
	return get_supplicant_config_value_list (wsconf, "proto", TRUE);
}

static GSList *
get_pairwise(GHashTable *wsconf)
{
	return get_supplicant_config_value_list (wsconf, "pairwise", TRUE);
}

static GSList *
get_group(GHashTable *wsconf)
{
	return get_supplicant_config_value_list (wsconf, "group", TRUE);
}

static char *
get_wep_key(GHashTable *wsconf, int n)
{
	char *key_str = NULL;
	char *val = NULL;
	char *wep_key = NULL;
	size_t len;

	key_str = g_strdup_printf("wep_key%d", n);

	val = get_supplicant_config_value (wsconf, key_str, FALSE);
	if(!val)
		goto out;

	len = strlen(val);

	if (len == 10 || len == 26) {
		/* Hexadecimal WEP key */
		char *p = val;

		while (*p) {
			if (!g_ascii_isxdigit (*p)) {
				PLUGIN_WARNING ("Invalid hexadecimal WEP key.");
				goto out;
			}
			p++;
		}
	} else if (len == 5 || len == 13) {
		/* ASCII passphrase */
		char *p = val;

		while (*p) {
			if (!isascii ((int) (*p))) {
				PLUGIN_WARNING ("Invalid ASCII WEP passphrase.");
				goto out;
			}
			p++;
		}
	} else {
		PLUGIN_WARNING ("Invalid WEP key length.");
		goto out;
	}

	wep_key = g_strdup(val);

out:

	g_free (key_str);
	g_free (val);

	return wep_key;
}

static char *
get_psk(GHashTable *wsconf)
{
	char *value;

	value = get_supplicant_config_value (wsconf, "psk", FALSE);
	if(!value)
		return NULL;
	if (strlen (value) > 64 || strlen (value) < 8) {
		PLUGIN_WARNING ("Error loading psk: not between 8 and 64 characters inclusive");
		g_free (value);
		return NULL;
	}

	return value;
}

static NMSetting *
make_wireless_setting (GHashTable *wsconf, const char *iface_path, const char *profile, GError **error)
{
	NMSettingWireless *s_wireless;
	char *str;
	enum wireless_mode mode;
	int mtu;
	GByteArray *array = NULL;
	iplink_info_t *info;

	s_wireless = NM_SETTING_WIRELESS (nm_setting_wireless_new ());

	str = get_ssid (wsconf);
	if (str) {
		gsize len = strlen (str);

		if (len > 0 && len <= 32) {
			GByteArray *ssid;

			ssid = g_byte_array_sized_new (len);
			g_byte_array_append (ssid, (const guint8 *) str, len);
			g_object_set (s_wireless, NM_SETTING_WIRELESS_SSID, ssid, NULL);
			g_byte_array_free (ssid, TRUE);
		} else {
			g_set_error (error, ifcfg_plugin_error_quark (), 0,
			             "Invalid SSID '%s' (size %zu not between 1 and 32 inclusive)",
			             str, len);
			g_free (str);
			goto error;
		}
		g_free (str);
	}

	mode = get_wireless_mode (wsconf);
	switch(mode) {
		case WL_MODE_INFRASTRUCTURE:
			g_object_set (s_wireless, NM_SETTING_WIRELESS_MODE, "infrastructure", NULL);
			break;
		case WL_MODE_ADHOC:
			g_object_set (s_wireless, NM_SETTING_WIRELESS_MODE, "adhoc", NULL);
			break;
		case WL_MODE_AP:
			g_object_set (s_wireless, NM_SETTING_WIRELESS_MODE, "ap", NULL);
			break;
		default:
			g_set_error (error, ifcfg_plugin_error_quark (), 0, "Invalid mode");
	}

	info = get_iplink_info(iface_path, profile, error);
	if(info) {
		if(get_mtu(info, &mtu)) {
			g_object_set (s_wireless, NM_SETTING_WIRELESS_MTU, mtu, NULL);
		}
		array = get_new_mac (info);
		if(array) {
			g_object_set (s_wireless, NM_SETTING_WIRELESS_CLONED_MAC_ADDRESS, array, NULL);
			g_byte_array_free (array, TRUE);
		}
	}

	iplink_info_free(info);

	array = get_bssid(wsconf);
	if(array) {
		g_object_set (s_wireless, NM_SETTING_WIRELESS_BSSID, array, NULL);
		g_byte_array_free (array, TRUE);
	}

	return NM_SETTING (s_wireless);

error:
	if (s_wireless)
		g_object_unref (s_wireless);
	return NULL;
}

static char *
get_printable_ssid (NMSetting *setting)
{
	const GByteArray *ssid;
	char *printable_ssid = NULL;

	ssid = nm_setting_wireless_get_ssid (NM_SETTING_WIRELESS (setting));
	if (ssid)
		printable_ssid = nm_utils_ssid_to_utf8 (ssid);

	return printable_ssid;
}

static void
set_proto_cb(gpointer data, gpointer user_data)
{
	NMSettingWirelessSecurity *s_wireless_sec = (NMSettingWirelessSecurity *)user_data;

	g_return_if_fail(data != NULL);
	g_return_if_fail(s_wireless_sec != NULL);

	/* NetworkManager don't understand 'WPA2', replace with 'RSN' */
	if (!g_ascii_strcasecmp ((const char *)data, "WPA2")) {
	    g_free(data);
	    data = g_strdup("rsn");
	}
	nm_setting_wireless_security_add_proto(s_wireless_sec, (const char *)data);
	g_free(data);
}

static void
set_pairwise_cb(gpointer data, gpointer user_data)
{
	NMSettingWirelessSecurity *s_wireless_sec = (NMSettingWirelessSecurity *)user_data;

	g_return_if_fail(data != NULL);
	g_return_if_fail(s_wireless_sec != NULL);

	nm_setting_wireless_security_add_pairwise(s_wireless_sec, (const char *)data);
	g_free(data);
}

static void
set_group_cb(gpointer data, gpointer user_data)
{
	NMSettingWirelessSecurity *s_wireless_sec = (NMSettingWirelessSecurity *)user_data;

	g_return_if_fail(data != NULL);
	g_return_if_fail(s_wireless_sec != NULL);

	nm_setting_wireless_security_add_group(s_wireless_sec, (const char *)data);
	g_free(data);
}

static NMSetting *
make_wireless_security_setting (GHashTable *wsconf, NMSettingWireless *s_wireless, GError **error)
{
	NMSettingWirelessSecurity *s_wireless_sec;
	char *auth_alg = get_auth_alg(wsconf);
	char *str;
	GSList *list;
	int i;

	/*FIXME: make_leap_settings() not implemented */
// 	if (auth_alg && !g_ascii_strcasecmp (str, "leap"))
// 		return make_leap_settings(wsconf, error);

	s_wireless_sec = NM_SETTING_WIRELESS_SECURITY (nm_setting_wireless_security_new ());

	str = get_key_mgmt(wsconf);
	if(str) {
		g_object_set (s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_KEY_MGMT, str, NULL);
		g_free(str);
	}

	g_object_set (s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_WEP_TX_KEYIDX, get_wep_tx_keyidx(wsconf), NULL);

	if(auth_alg) {
		g_object_set (s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_AUTH_ALG, auth_alg, NULL);
		g_free(auth_alg);
	}

	list = get_proto(wsconf);
	if(list) {
		g_slist_foreach(list, set_proto_cb, s_wireless_sec);
		g_slist_free(list);
	}

	list = get_pairwise(wsconf);
	if(list) {
		g_slist_foreach(list, set_pairwise_cb, s_wireless_sec);
		g_slist_free(list);
	}

	list = get_group(wsconf);
	if(list) {
		g_slist_foreach(list, set_group_cb, s_wireless_sec);
		g_slist_free(list);
	}

	for(i = 0; i <= 3; ++i) {
		str = get_wep_key(wsconf, i);
		if(str) {
			nm_setting_wireless_security_set_wep_key (s_wireless_sec, i, str);
			g_free(str);
		}
	}

	str = get_psk(wsconf);
	if(str) {
		g_object_set (s_wireless_sec, NM_SETTING_WIRELESS_SECURITY_PSK, str, NULL);
		g_free(str);
	}

	return (NMSetting *) s_wireless_sec;
}

static NMConnection *
read_wireless_connection (const char *iface_path, const char *profile, options_file_t *options,
				GHashTable *wsconf, gboolean *autoconnect, GError **error)
{
	NMConnection *connection = NULL;
	NMSetting *setting;
	NMSetting *security;
	char *printable_ssid;

	g_return_val_if_fail (iface_path != NULL, NULL);
	g_return_val_if_fail (options != NULL, NULL);
	g_return_val_if_fail (error != NULL, NULL);
	g_return_val_if_fail (*error == NULL, NULL);

	connection = nm_connection_new ();
	if (!connection) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Failed to allocate new connection for %s.", iface_path);
		return NULL;
	}

	if(!wsconf) {
		g_object_unref (connection);
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
					"Wireless settings are not found in %s.", iface_path);
		return NULL;
	}
	setting = make_wireless_setting (wsconf, iface_path, profile, error);
	if (!setting) {
		g_object_unref (connection);
		return NULL;
	}

	nm_connection_add_setting (connection, setting);
	security = make_wireless_security_setting (wsconf, NM_SETTING_WIRELESS (setting), error);
	if (security) {
		const char *security_str;
		if (NM_IS_SETTING_802_1X (security))
			security_str = NM_SETTING_802_1X_SETTING_NAME;
		else if (NM_IS_SETTING_WIRELESS_SECURITY (security))
			security_str = NM_SETTING_WIRELESS_SECURITY_SETTING_NAME;
		else {
			security_str = NULL;
			PLUGIN_WARNING ("Invalid security type: '%s'", G_OBJECT_TYPE_NAME (security));
		}
		g_object_set (G_OBJECT (setting), NM_SETTING_WIRELESS_SEC, security_str, NULL);
		nm_connection_add_setting (connection, security);
	}
	printable_ssid = get_printable_ssid (setting);
	setting = make_connection_setting (iface_path, profile, options,
						NM_SETTING_WIRELESS_SETTING_NAME, printable_ssid, autoconnect);
	nm_connection_add_setting (connection, setting);
	g_free (printable_ssid);

	return connection;
}

NMConnection *
read_connection (const char *iface_path, const char *profile, NMDeviceType dev_type, GHashTable *wsconf,
									gboolean *ignored, gboolean *autoconnect, GError **error)
{
	NMConnection *connection = NULL;
	options_file_t *options = NULL;
	char *nmc = NULL;
	NMSetting *s_ip4 = NULL, *s_ip6 = NULL;
	char *iface_name = NULL;
	gboolean config_ipv4, config_ipv6;

	g_return_val_if_fail (iface_path != NULL, NULL);
	g_return_val_if_fail (ignored != NULL, NULL);

	*ignored = FALSE;

	iface_name = get_iface_name (iface_path);
	if (!iface_name) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Invalid connection dir '%s'.", iface_path);
		goto done;
	}

	options = open_options_file(iface_path, profile);
	if (!options) {
		g_set_error (error, ifcfg_plugin_error_quark (), 0,
		             "Couldn't parse file '%s/options (%s)'", iface_path, profile ? profile : "default");
		goto done;
	}

	nmc = options_get_value (options, "NM_CONTROLLED");
	if (nmc) {
		nmc = g_strchomp (nmc);
		if (!g_ascii_strcasecmp("no", nmc) ||
				!g_ascii_strcasecmp("n", nmc) ||
				!g_ascii_strcasecmp("false", nmc))
			*ignored = TRUE;
		g_free (nmc);
	} else if (options->options) {
		/* if 'options' file for interface present and
		   NM_CONTROLLED not set then ignore this interface */
		*ignored = TRUE;
	}

	if (*ignored)
		goto done;

	config_ipv4 = options_true_value (options, "CONFIG_IPV4", TRUE) > 0;
	config_ipv6 = options_true_value (options, "CONFIG_IPV6", FALSE) > 0;

	/* If both CONFIG_IPV4 and CONFIG_IPV6 is FALSE
	 * ignore the interface */
	if (!config_ipv4 && !config_ipv6) {
		*ignored = TRUE;
		goto done;
	}

	switch (dev_type) {
	case NM_DEVICE_TYPE_ETHERNET:
		connection = read_wired_connection (iface_path, profile, options, autoconnect, error);
		break;
	case NM_DEVICE_TYPE_WIFI:
		connection = read_wireless_connection (iface_path, profile, options, wsconf, autoconnect, error);
		break;
	default:
		break;
	}

	if (!connection)
		goto done;

	if (config_ipv4) {
		s_ip4 = make_ip4_setting (options, iface_path, profile, error);
		if ( !s_ip4 ) {
			g_object_unref (connection);
			connection = NULL;
			goto done;
		} else {
			nm_connection_add_setting (connection, s_ip4);
		}
	}
	if (config_ipv6) {
		s_ip6 = make_ip6_setting (options, iface_path, profile, error);
		if ( !s_ip6 ) {
			g_object_unref (connection);
			connection = NULL;
			goto done;
		} else {
			nm_connection_add_setting (connection, s_ip6);
		}
	}

	if (!nm_connection_verify (connection, error)) {
		PLUGIN_WARNING ("%s: Connection for %s: '%s' / '%s' invalid: %d",
		           __func__, iface_name,
		           g_type_name (nm_connection_lookup_setting_type_by_quark ((*error)->domain)),
	           (*error)->message, (*error)->code);
		g_object_unref (connection);
		connection = NULL;
	}

done:
	g_free (iface_name);
	close_options_file(options);

	return connection;
}
